# -*- coding: utf-8 -*-
#
# ApplicationIntegrationServer - Provides an extensible Application 
# Integration Server for MailStore
#
# Copyright (c) 2014 - 2024 MailStore Software GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import os
import ssl
import sys
import json
import base64
import logging
import logging.handlers
import importlib
import http.server
import urllib.parse
import socketserver
import configparser


class Server(socketserver.ThreadingMixIn, http.server.HTTPServer):
    def __init__(self, *args, **kwargs):
        # Initialize HTTP server
        self.configuration = kwargs["configuration"]

        self.log = self.init_logging()

        self.log.warning("Starting Application Integration Server")

        # Load plugins
        self.log.info("SERVER loading plugins")
        self.plugins = self.init_plugins()

        self.log.warning(f"SERVER plugins used: {', '.join(self.plugins.keys())}")

        # load favicon for browser access
        with open("favicon.ico", mode="rb") as iconfile:
            self.icon = iconfile.read()

        # Remove "configuration" so **kwargs can be passed to http.server.HTTPServer.__init__()
        del kwargs["configuration"]

        # Start HTTP server
        http.server.HTTPServer.__init__(self, *args, **kwargs)
        self.log.warning(f"SERVER listening on {self.server_address[0]}:{self.server_address[1]}")

        # Initialize TLS, if enabled
        self.init_tls()

    def init_logging(self):
        """Initialize logging"""
        logger = logging.getLogger("AppIntLogger")
        logger.setLevel(logging.DEBUG)
        # logger.propagate = False
        formatter = logging.Formatter("%(asctime)s %(levelname)-8s %(module)s.%(funcName)s() [%(lineno)d]: %(message)s")

        # Set up logging to standard out if it is enabled
        if self.configuration["log"].getboolean("stdout"):
            streamhandler = logging.StreamHandler(sys.stdout)
            config_stdout_level = self.configuration["log"]["stdout_level"]
            streamhandler.setLevel(getattr(logging, config_stdout_level))
            streamhandler.setFormatter(formatter)
            logger.addHandler(streamhandler)

        # Set up logging to file system if it is enabled
        if self.configuration["log"].getboolean("file"):
            directory = self.configuration["log"]["directory"]
            os.makedirs(directory, exist_ok=True)

            debug_log_file = os.path.join(directory, "AppInt.log")
            filehandler = logging.handlers.TimedRotatingFileHandler(debug_log_file, when="D", interval=1,
                                                                    backupCount=self.configuration["log"].getint(
                                                                        "days"))
            config_file_level = self.configuration["log"]["file_level"]
            filehandler.setLevel(getattr(logging, config_file_level))
            filehandler.setFormatter(formatter)
            logger.addHandler(filehandler)

        return logger

    def init_plugins(self, plugin_dir="plugins"):
        """Initialize plugins"""
        plugin_dir_absolute = os.path.join(os.path.dirname(os.path.abspath(__file__)), plugin_dir)
        self.log.info(f"PLUGIN directory: {plugin_dir_absolute}")

        # Load the plugin classes into a dictionary, where the key is the filename.
        # This allows us to call the plugin methods as follows:
        #     self.server.plugins[pluginname].list_() or self.server.plugins[pluginname].auth(username, password)
        # We can pass the "path" of the HTTP command as parameter to the 'server.plugins' dictionary
        plugins = dict()
        for file in os.listdir(plugin_dir_absolute):
            self.log.debug(f"PLUGIN {file} found")

            if not os.path.isfile(os.path.join(plugin_dir_absolute, file)):
                self.log.info(f"PLUGIN {file} ignored, as it is not a file")
                continue
            if not file.endswith(".py"):
                self.log.info(f"PLUGIN {file} ignored, as it is not a python file")
                continue
            if file.startswith("_"):
                self.log.info(f"PLUGIN {file} ignored, file does not meet naming convention")
                continue

            # Remove file extension
            plugin_file = os.path.splitext(file)[0]
            self.log.info(f"PLUGIN {plugin_file} importing")
            try:
                # Import plugin file as module
                plugin_import_object = importlib.import_module(f"{plugin_dir}.{plugin_file}")
            except Exception as e:
                self.log.error(f"PLUGIN {plugin_file} ERROR while importing: {repr(e)}")
            else:
                self.log.debug(f"PLUGIN {plugin_file} imported successfully, loading 'Plugin' class")
                try:
                    # Instantiate Plugin() class
                    plugin_class = plugin_import_object.Plugin()
                except Exception as e:
                    self.log.error(f"PLUGIN {plugin_file}.Plugin() class cannot be loaded: {repr(e)}")
                else:
                    self.log.debug(f"PLUGIN {plugin_file} loaded successfully")

                    # Check if plugin has all necessary properties
                    if not hasattr(plugin_class, "auth"):
                        self.log.error(f"PLUGIN {plugin_file} cannot be used: 'auth' method is missing")
                        continue
                    if not hasattr(plugin_class, "list_"):
                        self.log.error(f"PLUGIN {plugin_file} cannot be used: 'list_' method is missing")
                        continue

                    # Check if plugin requires authentication
                    if hasattr(plugin_class, "http_authentication"):
                        self.log.info(f"PLUGIN {plugin_file} requires HTTP authentication")

                    # Add plugin to global list
                    plugins[plugin_file] = plugin_class
                    self.log.info(f"PLUGIN {plugin_file} imported successfully")

        return plugins

    def init_tls(self):
        """Initialize TLS encryption"""

        # Load certificate and initialize TLS only if it is enabled in the configuration
        if self.configuration["tls"].getboolean("enabled"):
            # Convert relative path to certificate file into absolute path...
            cert_file_absolute = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                              self.configuration["tls"]["certificate"])

            # ...otherwise assume absolute path was already given.
            if not os.path.exists(cert_file_absolute):
                cert_file_absolute = self.configuration["tls"]["certificate"]

            # When a certificate could be found, wrap the socket
            if os.path.exists(cert_file_absolute):
                self.log.info(f"SERVER TLS certificate: {cert_file_absolute}")
                ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
                ssl_context.load_cert_chain(certfile=cert_file_absolute)
                if sys.version_info >= (3, 7):
                    ssl_context.minimum_version = ssl.TLSVersion.TLSv1_2
                self.socket = ssl_context.wrap_socket(self.socket, server_side=True)
                self.log.info("SERVER TLS activated")
            else:
                message = f"SERVER TLS certificate {cert_file_absolute} cannot be not found, server will not be started"
                self.log.critical(message)
                raise Exception(message)

        # TLS is disabled in the configuration
        else:
            self.log.warning("SERVER TLS is disabled in global configuration")


class RequestHandler(http.server.BaseHTTPRequestHandler):
    server: Server
    error_template = "<!DOCTYPE html><head><title>MailStore Application Integration Server</title></head><body>{}</body></html>"

    def do_GET(self):
        """Handle HTTP GET requests by parsing the query string parameters"""
        raw = urllib.parse.urlparse(self.path).query
        self.server.log.debug(f"{self.client_address[0]}:{self.client_address[1]} GET data: {raw}")
        data = urllib.parse.parse_qs(raw)
        self.handle_request(data)

    def do_POST(self):
        """Handle HTTP POST requests by parsing the POST body"""
        content_length = int(self.headers["Content-Length"])
        raw_body = self.rfile.read(content_length).decode()
        self.server.log.debug(f"{self.client_address[0]}:{self.client_address[1]} POST data: {raw_body}")
        data = urllib.parse.parse_qs(raw_body)
        # Also handle query parameter
        raw_query = urllib.parse.urlparse(self.path).query
        self.server.log.debug(f"{self.client_address[0]}:{self.client_address[1]} QUERY data: {raw_query}")
        data.update(urllib.parse.parse_qs(raw_query))
        self.handle_request(data)

    def log_message(self, format_, *args):
        """This method overwrites the default method to suppress writing to STDERR,
        which fails when running as a service, making the service completely unusable.
        See https://stackoverflow.com/questions/10651052/how-to-quiet-simplehttpserver for details."""
        self.server.log.debug(f"{self.client_address[0]}:{self.client_address[1]} {format_ % args}")

    def is_http_user_authenticated(self, plugin):
        """If a plugin requires additional authentication, try to authenticate the user."""

        # Enable to show passwords in DEBUG output
        show_passwords = False

        # No authentication required by plugin
        if not hasattr(self.server.plugins[plugin], "http_authentication"):
            self.server.log.debug(f"{self.client_address[0]}:{self.client_address[1]} NO authentication required")
            return True

        # Authentication required by plugin
        self.server.log.debug(f"{self.client_address[0]}:{self.client_address[1]} authentication required")
        authorize_header = self.headers["authorization"]

        # Authorization required, but no header is sent by the client
        if not authorize_header:
            self.server.log.debug(f"{self.client_address[0]}:{self.client_address[1]} NO auth header sent by client")
            return False

        if show_passwords:
            self.server.log.debug(f"{self.client_address[0]}:{self.client_address[1]} auth header: {authorize_header}")

        # Usually in the format: Basic base64(username:password)
        # type_ (Basic) will not be used
        type_, value = authorize_header.split()
        decoded_value = base64.b64decode(value.encode()).decode()
        if ":" in decoded_value:
            username, password = decoded_value.split(":", 1)
        else:
            self.server.log.debug(
                f"{self.client_address[0]}:{self.client_address[1]} invalid auth header sent by client")
            return False

        # Send http username and password to plugin
        is_authenticated = self.server.plugins[plugin].http_authentication(username, password)

        self.server.log.debug(f"{self.client_address[0]}:{self.client_address[1]} "
                              f"credentials: '{username}' '{password if show_passwords else '<hidden>'}', "
                              f"is authenticated: {is_authenticated}")
        return is_authenticated

    def handle_request(self, data):
        """Handle parsed request data from do_GET and do_POST methods.
        
        data: Dictionary of request parameters
        """

        self.server.log.debug(f"{self.client_address[0]}:{self.client_address[1]} line: {self.requestline}")
        path = urllib.parse.urlparse(self.path).path
        self.server.log.debug(f"{self.client_address[0]}:{self.client_address[1]} path: {path}")
        self.server.log.debug(f"{self.client_address[0]}:{self.client_address[1]} data: {data}")
        path = path.strip("/")

        if path == "favicon.ico":
            self.send_icon()
            return

        # Path must match a plugin name, else return HTTP 404
        if path not in self.server.plugins.keys():
            self.server.log.warning(
                f"{self.client_address[0]}:{self.client_address[1]} HTTP 404 Plugin '{path}' not found")
            self.send_404(path)
            return

        # Check if http user is authenticated, else return HTTP 401
        if not self.is_http_user_authenticated(path):
            self.server.log.warning(f"{self.client_address[0]}:{self.client_address[1]} HTTP 401 authentication required for '{path}'")
            self.send_401()
            return

        # Parameter 'cmd' must always be present, else return HTTP 400
        if "cmd" not in data:
            message = "'cmd' parameter missing."
            self.server.log.warning(f"{self.client_address[0]}:{self.client_address[1]} HTTP 400 {message}")
            self.send_400(message)
            return

        # List users
        if data["cmd"][0] == "list":
            try:
                # Execute list_() method from plugin that matches path name
                user_list = self.server.plugins[path].list_(data)
            except Exception as e:
                self.server.log.error(f"{self.client_address[0]}:{self.client_address[1]} HTTP 500 user list of '{path}': {repr(e)}")
                # Don't send 'e' to the client, because sensitive information could leak
                self.send_500()
            else:
                self.server.log.info(f"{self.client_address[0]}:{self.client_address[1]} HTTP 200 user list of '{path}'")
                try:
                    self.send_200(user_list)
                except Exception as e:
                    self.server.log.error(f"SERVER ERROR while sending HTTP 200 response to list_: {repr(e)}")
                else:
                    self.server.log.debug("SERVER HTTP 200 OK")
        # Authenticate a user
        elif data["cmd"][0] == "auth":
            # Parameters 'user' and 'pass' must be given, else return HTTP 400
            message = ""
            if "user" not in data:
                message = "'user' parameter missing."
            if "pass" not in data:
                if message:
                    message += "<br/>"
                message += "'pass' parameter missing."

            if message:
                self.server.log.warning(f"{self.client_address[0]}:{self.client_address[1]} HTTP 400 {message}")
                self.send_400(message)
                return

            # Execute auth() method from plugin that matches path name
            try:
                user_is_authenticated = self.server.plugins[path].auth(data["user"][0], data["pass"][0])
            except Exception as e:
                self.server.log.error(f"{self.client_address[0]}:{self.client_address[1]} HTTP 500 auth of '{data['user'][0]}': {repr(e)}")
                # Don't send 'e' to the client, because sensitive information could leak
                self.send_500()
            else:
                if isinstance(user_is_authenticated, bool):
                    self.server.log.warning(
                        f"PLUGIN '{path}' returns 'bool' type instead of a 'dict' type to an 'auth' call."
                        "performing conversion")
                    user_is_authenticated = {"succeeded": user_is_authenticated}
                self.server.log.info(f"{self.client_address[0]}:{self.client_address[1]} HTTP 200 auth: {data['user'][0]}@{path} {str(user_is_authenticated)}")
                try:
                    self.send_200(user_is_authenticated)
                except Exception as e:
                    self.server.log.error(f"SERVER ERROR while sending HTTP 200 response to auth: {repr(e)}")
                else:
                    self.server.log.debug("SERVER HTTP 200 OK")
        # No valid cmd parameter value
        else:
            message = f"unknown command: {data['cmd'][0]}."
            self.server.log.warning(f"{self.client_address[0]}:{self.client_address[1]} HTTP 400 {message}")
            self.send_400(message)

    def send_icon(self):
        """Send favicon
        """
        self.send_response_only(200)
        self.send_header("Content-Type", "image/x-icon")
        self.send_header("Content-Length", str(len(self.server.icon)))
        self.end_headers()
        self.server.log.debug(f"SERVER sending favicon of length: {len(self.server.icon)}")
        self.wfile.write(self.server.icon)

    def send_200(self, data):
        """Send HTTP 200 OK response

        data: Response body
        """
        data = json.dumps(data).encode()
        self.send_response_only(200)
        self.send_header("Content-Type", "application/json; charset=utf-8")
        self.send_header("Content-Length", str(len(data)))
        self.end_headers()
        self.server.log.debug(f"SERVER sending data: {data.decode()}")
        self.wfile.write(data)

    def send_400(self, message=""):
        """Send HTTP 400 Bad request"""
        error_message = RequestHandler.error_template.format(f"400: Bad request.<br/>{message}").encode()
        self.send_response_only(400)
        self.send_header("Content-type", "text/html; charset=utf-8")
        self.send_header("Content-Length", str(len(error_message)))
        self.end_headers()
        self.wfile.write(error_message)

    def send_401(self):
        """Send HTTP 401 Authentication required"""
        error_message = RequestHandler.error_template.format("401: Authentication required.").encode()
        self.send_response_only(401)
        self.send_header("www-authenticate", 'Basic Realm=""')
        self.send_header("Content-type", "text/html; charset=utf-8")
        self.send_header("Content-Length", str(len(error_message)))
        self.end_headers()
        self.wfile.write(error_message)

    def send_404(self, path):
        """Send HTTP 404 Plugin not found
        
        path: Path representing the plugin name
        """
        error_message = RequestHandler.error_template.format(f"404: Plugin '{path}' not found.").encode()
        self.send_response_only(404)
        self.send_header("Content-type", "text/html; charset=utf-8")
        self.send_header("Content-Length", str(len(error_message)))
        self.end_headers()
        self.wfile.write(error_message)

    def send_500(self):
        """Send HTTP 500 Internal Server Error"""
        error_message = RequestHandler.error_template.format("500: Internal Server Error").encode()
        self.send_response_only(500)
        self.send_header("Content-type", "text/html; charset=utf-8")
        self.send_header("Content-Length", str(len(error_message)))
        self.end_headers()
        self.wfile.write(error_message)


def main():
    # Read configuration from 'config.ini' file
    configuration = configparser.ConfigParser()
    configuration.read(os.path.join(os.path.dirname(os.path.abspath(__file__)), "config.ini"))
    ipaddress = configuration["network"]["ipaddress"]
    port = configuration["network"].getint("port")

    try:
        # Initialize the server
        server = Server((ipaddress, port), RequestHandler, configuration=configuration)
    except Exception as ex:
        print(repr(ex))
    else:
        # Start server if initialization was successful
        server.serve_forever()


if __name__ == '__main__':
    main()
